//---------------------------------------------------------------------------
// Copyright (C) 2021 Bo Xu <xubo123@big.ac.cn>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Read mutations from VCF file and/or convert VCF into mutation format file
//     The mutation format:
//         <SampleName>\t<AccessionID>\t[<POS>(<VariantType>:<REF>-><ALT>)[;<POS>(<VariantType>:<REF>-><ALT>)]...]
//---------------------------------------------------------------------------

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <string.h>
#include "optionparser.h"

#if defined(_WIN32)
#include <io.h>

#define IS_STDIN_REDIRECTED (!(_isatty(_fileno(stdin))))

#elif defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__))
#include <unistd.h>

#define IS_STDIN_REDIRECTED (!(isatty(fileno(stdin))))

#else
#define IS_STDIN_REDIRECTED 0
#endif

extern "C" {
    void version();
    int main_vcf2mut(int argc, char* argv[]);
    typedef int(*Callback)(const char*, size_t, size_t, size_t, void*);
    int vcf2mut(const char* vcfName, const char* mutName, Callback lineHandler, void* arg, const char* missingValue);
}

namespace VCF2mut {

void help() {
    printf("Usage: McAN vcf2mut [option]...\n");
    printf("       McAN vcf2mut [<vcf> [mutation]]\n\n");

    printf("Options:\n");
    printf("  -f, --vcf <file>              input vcf file [default: stdin]\n");
    printf("  -u, --mut <file>              output mutation file [default: stdout]\n\n");

    printf("  -h, --help                    display this help and exit\n");
    printf("  -v, --version                 output version information and exit\n");
}

enum EOption {
    Evcf = 0,
    Emut,
    Eversion,
    Ehelp,
    ELast
};

struct Argument{
    const char* vcf;
    const char* mut;
    bool version;
    bool help;
    bool used[ELast];
};

bool getOptions(int argc, char* argv[], Argument* arg) {
    if (argc <= 1) {
        help();
        return false;
    }

    static const Option options[] = {
        { "--vcf", 1, Evcf },
        { "--mut", 1, Emut },
        { "--version", 0, Eversion },
        { "--help", 0, Ehelp },
        { "-f", 1, Evcf },
        { "-u", 1, Emut },
        { "-v", 0, Eversion },
        { "-h", 0, Ehelp },
        { "", 1, Evcf },
        { "", 1, Emut },
    };

    return ArgParser::parseCommandLine(argc, argv, options, LENGTH_OF(options), 1, 1, [arg](Arg a) {
        int i = a.id;
        const char* val = a.val;

        if (arg->used[i])
            return ERepeated;
        arg->used[i] = true;

        switch (i)
        {
        case Evcf:
            arg->vcf = val;
            break;
        case Emut:
            arg->mut = val;
            break;
        case Eversion:
            version();
            return ESuccessEnd;
        case Ehelp:
            help();
            return ESuccessEnd;
        default:
            return EInvalid;
        }

        return ESuccess;
    });
}

}

int main_vcf2mut(int argc, char* argv[]) {
    using namespace VCF2mut;

    Argument arg;
    memset(&arg, 0, sizeof(arg));
    if (!getOptions(argc, argv, &arg))
        return EXIT_FAILURE;
    else if (arg.used[Ehelp] || arg.used[Eversion])
        return EXIT_SUCCESS;

    if (!arg.used[Evcf] && !IS_STDIN_REDIRECTED) {
        fprintf(stderr, "Error: input file is required.\n");
        return EXIT_FAILURE;
    }

    if (!arg.used[Evcf])
        arg.vcf = "-";

    if (!arg.used[Emut])
        arg.mut = "-";

    if (-1 == vcf2mut(arg.vcf, arg.mut, NULL, NULL, "*"))
        return EXIT_FAILURE;
    else
        return EXIT_SUCCESS;
}
