//---------------------------------------------------------------------------
// Copyright (C) 2021 Bo Xu <xubo123@big.ac.cn>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Command line options parser
//---------------------------------------------------------------------------

#include <cstring>
#include <iostream>
#include "optionparser.h"

namespace ArgParser {

bool parseArg(int argc, char* const argv[], const Option options[], int size, Opt& opt, int& argi, IntSt& st, bool printErrMsg, bool POSIXLY_CORRECT) {
    if (st.optIdx >= argc) {
        argi = argc;
        return false;
    }

    bool ret = true;

    if (!st.stopOptParsing && argv[st.optIdx][0] == '-' && argv[st.optIdx][1] != '-' && argv[st.optIdx][1] != 0) {
        int i;
        for (i = 0; i < size; i++) {
            if (options[i].name[0] == '-' && options[i].name[1] == argv[st.optIdx][st.charIdx] && options[i].name[2] == 0)
                break;
        }
        if (i < size) {
            if (options[i].hasArg) {
                if (argv[st.optIdx][st.charIdx + 1] != 0) {
                    opt = { i, argv[st.optIdx] + st.charIdx + 1 };
                    argi = ++st.optIdx;
                    st.charIdx = 1;
                    ret = true;
                }
                else if (st.optIdx + 1 < argc && argv[st.optIdx + 1][0] != 0) {
                    opt = { i, argv[++st.optIdx] };
                    argi = ++st.optIdx;
                    st.charIdx = 1;
                    ret = true;
                }
                else {
                    if (printErrMsg)
                        std::cerr << "Error: option '" << options[i].name << "' requires an argument.\n";
                    argi = st.optIdx;
                    ret = false;
                }
            }
            else {
                opt = { i, nullptr };
                if (argv[st.optIdx][++st.charIdx] == 0) {
                    st.optIdx++;
                    st.charIdx = 1;
                }
                argi = st.optIdx;
                ret = true;
            }
        }
        else {
            if (printErrMsg)
                std::cerr << "Error: unrecognized option '" << '-' << argv[st.optIdx][st.charIdx] << "'\n";
            argi = st.optIdx;
            ret = false;
        }
    }
    else if (!st.stopOptParsing && argv[st.optIdx][0] == '-' && argv[st.optIdx][1] == '-' && argv[st.optIdx][2] != '=' && argv[st.optIdx][2] != 0) {
        std::size_t optSize = std::strlen(argv[st.optIdx]);
        std::size_t iOptSize;
        int i;
        for (i = 0; i < size; i++) {
            iOptSize = std::strlen(options[i].name);
            if (optSize >= iOptSize && std::memcmp(argv[st.optIdx], options[i].name, iOptSize) == 0 && (argv[st.optIdx][iOptSize] == 0 || argv[st.optIdx][iOptSize] == '='))
                break;
        }
        if (i < size) {
            if (options[i].hasArg) {
                if (argv[st.optIdx][iOptSize] == '=' && argv[st.optIdx][iOptSize + 1] != 0) {
                    opt = { i, argv[st.optIdx] + iOptSize + 1 };
                    argi = ++st.optIdx;
                    ret = true;
                }
                else if (argv[st.optIdx][iOptSize] == 0 && st.optIdx + 1 < argc && argv[st.optIdx + 1][0] != 0) {
                    opt = { i, argv[++st.optIdx] };
                    argi = ++st.optIdx;
                    ret = true;
                }
                else {
                    if (printErrMsg)
                        std::cerr << "Error: option '" << options[i].name << "' requires an argument.\n";
                    argi = st.optIdx;
                    ret = false;
                }
            }
            else {
                if (argv[st.optIdx][iOptSize] == 0) {
                    opt = { i, nullptr };
                    argi = ++st.optIdx;
                    ret = true;
                }
                else {
                    if (printErrMsg)
                        std::cerr << "Error: option '" << options[i].name << "' doesn't allow an argument.\n";
                    argi = st.optIdx;
                    ret = false;
                }
            }
        }
        else {
            if (printErrMsg)
                std::cerr << "Error: unrecognized option '" << argv[st.optIdx] << "'\n";
            argi = st.optIdx;
            ret = false;
        }
    }
    else {
        if (!st.stopOptParsing) {
            if (argv[st.optIdx][0] == '-' && argv[st.optIdx][1] == '-' && argv[st.optIdx][2] == 0) {
                st.stopOptParsing = true;
                ++st.optIdx;
                if (st.optIdx >= argc) {
                    argi = argc;
                    ret = false;
                }
            }
            else if (POSIXLY_CORRECT)
                st.stopOptParsing = true;
        }

        if (ret) {
            int i;
            for (i = st.nonOptIdx; i < size; i++) {
                if (options[i].name[0] == 0)
                    break;
            }
            if (i < size) {
                opt = { i, argv[st.optIdx] };
                argi = ++st.optIdx;
                st.nonOptIdx = i + 1;
                ret = true;
            }
            else {
                if (printErrMsg)
                    std::cerr << "Error: unrecognized argument '" << argv[st.optIdx] << "'\n";
                argi = st.optIdx;
                ret = false;
            }
        }
    }

    return ret;
}

bool handleError(int code, bool printErrMsg, const char* name, const char* val) {
    bool ret = true;

    if (code == ERepeated) {
        if (printErrMsg) {
            if (name[0] == 0)
                std::cerr << "Error: repeated argument '" << val << "'\n";
            else
                std::cerr << "Error: repeated option '" << name << "'\n";
        }
        ret = false;
    }
    else if (code == EInvalid) {
        if (printErrMsg) {
            if (name[0] == 0)
                std::cerr << "Error: invalid argument '" << val << "'\n";
            else
                std::cerr << "Error: invalid '" << name << "' argument '" << val << "'\n";
        }
        ret = false;
    }

    return ret;
}

};

int parseCommandLine(int argc, char* const argv[], const Option options[], int size, int printErrMsg, int POSIXLY_CORRECT, Parser f, void* args) {
    return ArgParser::parseCommandLine(argc, argv, options, size, !!printErrMsg, !!POSIXLY_CORRECT, f, args);
}
